@@ -14,40 +14,61 @@
 #include <linux/device.h>
 #include <linux/dma-mapping.h>
 #include <linux/slab.h>
-#include <linux/sys_soc.h>
 
 #include <media/media-entity.h>
-#include <media/rcar-fcp.h>
 #include <media/v4l2-subdev.h>
 #include <media/vsp1.h>
 
 #include "vsp1.h"
 #include "vsp1_bru.h"
-#include "vsp1_brs.h"
 #include "vsp1_dl.h"
 #include "vsp1_drm.h"
 #include "vsp1_lif.h"
 #include "vsp1_pipe.h"
 #include "vsp1_rwpf.h"
-#include "vsp1_video.h"
 
 
 /* -----------------------------------------------------------------------------
  * Interrupt Handling
  */
 
-void vsp1_drm_display_start(struct vsp1_device *vsp1, unsigned int lif_index)
+void vsp1_drm_display_start(struct vsp1_device *vsp1, unsigned int pipe_index,
+			    struct vsp1_pipeline *pipe)
 {
-	vsp1_dlm_irq_display_start(vsp1->drm->pipe[lif_index].output->dlm);
+	if (pipe->output->write_back == WB_STAT_CATP_REQUEST) {
+		pipe->output->write_back = WB_STAT_CATP_SET;
+		wake_up_interruptible(&pipe->event_wait);
+	} else if (pipe->completed) {
+		bool writeback;
+		u32 offset = 0x100 * pipe_index;
+
+		if ((vsp1_read(vsp1, VI6_WPF_WRBCK_CTRL + offset) &
+		    VI6_WPF_WRBCK_CTRL_WBMD) == VI6_WPF_WRBCK_CTRL_WBMD)
+			writeback = true;
+		else
+			writeback = false;
+
+		if (pipe->output->write_back == WB_STAT_CATP_SET &&
+		    writeback) {
+			pipe->output->write_back = WB_STAT_CATP_START;
+			wake_up_interruptible(&pipe->event_wait);
+		} else if (pipe->output->write_back == WB_STAT_CATP_START &&
+			   !writeback) {
+			pipe->output->write_back = WB_STAT_CATP_DONE;
+			wake_up_interruptible(&pipe->event_wait);
+		}
+	}
+
+	pipe->completed = false;
 }
 
 static void vsp1_du_pipeline_frame_end(struct vsp1_pipeline *pipe,
-				       unsigned int lif_index)
+				       bool completed)
 {
-	struct vsp1_drm *drm = to_vsp1_drm(pipe, lif_index);
+	struct vsp1_drm_pipeline *drm_pipe = to_vsp1_drm_pipeline(pipe);
 
-	if (drm->du_complete[lif_index])
-		drm->du_complete[lif_index](drm->du_private[lif_index]);
+	if (drm_pipe->du_complete)
+		drm_pipe->du_complete(drm_pipe->du_private, completed);
 }
 
 /* -----------------------------------------------------------------------------
@@ -65,10 +86,14 @@ int vsp1_du_init(struct device *dev)
 }
 EXPORT_SYMBOL_GPL(vsp1_du_init);
 
-int vsp1_du_if_set_mute(struct device *dev, bool on, unsigned int lif_index)
+int vsp1_du_if_set_mute(struct device *dev, bool on, unsigned int pipe_index)
 {
 	struct vsp1_device *vsp1 = dev_get_drvdata(dev);
-	struct vsp1_pipeline *pipe = &vsp1->drm->pipe[lif_index];
+	struct vsp1_drm_pipeline *drm_pipe;
+	struct vsp1_pipeline *pipe;
+
+	drm_pipe = &vsp1->drm->pipe[pipe_index];
+	pipe = &drm_pipe->pipe;
 
 	if (on)
 		pipe->vmute_flag = true;
@@ -82,49 +107,43 @@ EXPORT_SYMBOL_GPL(vsp1_du_if_set_mute);
 /**
  * vsp1_du_setup_lif - Setup the output part of the VSP pipeline
  * @dev: the VSP device
+ * @pipe_index: the DRM pipeline index
  * @cfg: the LIF configuration
  *
  * Configure the output part of VSP DRM pipeline for the given frame @cfg.width
- * and @cfg.height. This sets up formats on the BRU source pad, the WPF0 sink
- * and source pads, and the LIF sink pad.
+ * and @cfg.height. This sets up formats on the blend unit (BRU or BRS) source
+ * pad, the WPF sink and source pads, and the LIF sink pad.
  *
- * As the media bus code on the BRU source pad is conditioned by the
- * configuration of the BRU sink 0 pad, we also set up the formats on all BRU
+ * The @pipe_index argument selects which DRM pipeline to setup. The number of
+ * available pipelines depend on the VSP instance.
+ *
+ * As the media bus code on the blend unit source pad is conditioned by the
+ * configuration of its sink 0 pad, we also set up the formats on all blend unit
  * sinks, even if the configuration will be overwritten later by
- * vsp1_du_setup_rpf(). This ensures that the BRU configuration is set to a well
- * defined state.
+ * vsp1_du_setup_rpf(). This ensures that the blend unit configuration is set to
+ * a well defined state.
  *
  * Return 0 on success or a negative error code on failure.
  */
-int vsp1_du_setup_lif(struct device *dev, const struct vsp1_du_lif_config *cfg,
-		      unsigned int lif_index, bool suspend)
+int vsp1_du_setup_lif(struct device *dev, unsigned int pipe_index,
+		      const struct vsp1_du_lif_config *cfg)
 {
 	struct vsp1_device *vsp1 = dev_get_drvdata(dev);
-	struct vsp1_pipeline *pipe = &vsp1->drm->pipe[lif_index];
-	struct vsp1_bru *bru = vsp1->bru;
-	struct vsp1_brs *brs = vsp1->brs;
-	unsigned int init_bru_num, end_bru_num;
-	unsigned int init_brs_num, end_brs_num;
+	struct vsp1_drm_pipeline *drm_pipe;
+	struct vsp1_pipeline *pipe;
+	struct vsp1_bru *bru;
 	struct v4l2_subdev_format format;
+	const char *bru_name;
 	unsigned int i;
 	int ret;
 
-	if (vsp1_gen3_vspdl_check(vsp1)) {
-		if (!vsp1->brs || !vsp1->lif[1])
-			return -ENXIO;
+	if (pipe_index >= vsp1->info->lif_count)
+		return -EINVAL;
 
-		init_bru_num = 0;
-		init_brs_num = vsp1->info->rpf_count -
-			       vsp1->num_brs_inputs;
-		end_bru_num = vsp1->info->rpf_count -
-			      vsp1->num_brs_inputs;
-		end_brs_num = vsp1->info->rpf_count;
-	} else {
-		init_bru_num = 0;
-		init_brs_num = 0;
-		end_bru_num = bru->entity.source_pad;
-		end_brs_num = 0;
-	}
+	drm_pipe = &vsp1->drm->pipe[pipe_index];
+	pipe = &drm_pipe->pipe;
+	bru = to_bru(&pipe->bru->subdev);
+	bru_name = pipe->bru->type == VSP1_ENTITY_BRU ? "BRU" : "BRS";
 
 	if (!cfg) {
 		/*
@@ -135,26 +154,27 @@ int vsp1_du_setup_lif(struct device *dev, const struct vsp1_du_lif_config *cfg,
 		if (ret == -ETIMEDOUT)
 			dev_err(vsp1->dev, "DRM pipeline stop timeout\n");
 
-		media_entity_pipeline_stop(&pipe->output->entity.subdev.entity);
+		media_pipeline_stop(&pipe->output->entity.subdev.entity);
 
-		if (!suspend) {
-			if (lif_index == 1) {
-				for (i = init_brs_num; i < end_brs_num; ++i) {
-					vsp1->drm->inputs[i].enabled = false;
-					brs->inputs[i].rpf = NULL;
-					pipe->inputs[i] = NULL;
-				}
-			} else {
-				for (i = init_bru_num; i < end_bru_num; ++i) {
-					vsp1->drm->inputs[i].enabled = false;
-					bru->inputs[i].rpf = NULL;
-					pipe->inputs[i] = NULL;
-				}
-			}
+		for (i = 0; i < ARRAY_SIZE(pipe->inputs); ++i) {
+			struct vsp1_rwpf *rpf = pipe->inputs[i];
+
+			if (!rpf)
+				continue;
+
+			/*
+			 * Remove the RPF from the pipe and the list of BRU
+			 * inputs.
+			 */
+			WARN_ON(list_empty(&rpf->entity.list_pipe));
+			list_del_init(&rpf->entity.list_pipe);
+			pipe->inputs[i] = NULL;
+
+			bru->inputs[rpf->bru_input].rpf = NULL;
 		}
 
+		drm_pipe->du_complete = NULL;
 		pipe->num_inputs = 0;
-		vsp1->drm->du_complete[lif_index] = NULL;
 
 		vsp1_dlm_reset(pipe->output->dlm);
 		vsp1_device_put(vsp1);
@@ -164,8 +184,8 @@ int vsp1_du_setup_lif(struct device *dev, const struct vsp1_du_lif_config *cfg,
 		return 0;
 	}
 
-	dev_dbg(vsp1->dev, "%s: configuring LIF%d with format %ux%u\n",
-		__func__, lif_index, cfg->width, cfg->height);
+	dev_dbg(vsp1->dev, "%s: configuring LIF%u with format %ux%u\n",
+		__func__, pipe_index, cfg->width, cfg->height);
 
 	/*
 	 * Configure the format at the BRU sinks and propagate it through the
@@ -174,106 +194,71 @@ int vsp1_du_setup_lif(struct device *dev, const struct vsp1_du_lif_config *cfg,
 	memset(&format, 0, sizeof(format));
 	format.which = V4L2_SUBDEV_FORMAT_ACTIVE;
 
-	if (lif_index == 1) {
-		for (i = init_brs_num; i < end_brs_num; ++i) {
-			format.pad = i;
+	for (i = 0; i < pipe->bru->source_pad; ++i) {
+		format.pad = i;
 
-			format.format.width = cfg->width;
-			format.format.height = cfg->height;
-			format.format.code = MEDIA_BUS_FMT_ARGB8888_1X32;
-			format.format.field = V4L2_FIELD_NONE;
+		format.format.width = cfg->width;
+		format.format.height = cfg->height;
+		format.format.code = MEDIA_BUS_FMT_ARGB8888_1X32;
+		format.format.field = V4L2_FIELD_NONE;
 
-			ret = v4l2_subdev_call(&brs->entity.subdev, pad,
-					       set_fmt, NULL, &format);
-			if (ret < 0)
-				return ret;
+		ret = v4l2_subdev_call(&pipe->bru->subdev, pad,
+				       set_fmt, NULL, &format);
+		if (ret < 0)
+			return ret;
 
-			dev_dbg(vsp1->dev,
-				"%s: set format %ux%u (%x) on BRS pad %u\n",
-				__func__, format.format.width,
-				format.format.height,
-				format.format.code, i);
-		}
-		format.pad = brs->entity.source_pad;
-	} else {
-		for (i = init_bru_num; i < end_bru_num; ++i) {
-			format.pad = i;
-
-			format.format.width = cfg->width;
-			format.format.height = cfg->height;
-			format.format.code = MEDIA_BUS_FMT_ARGB8888_1X32;
-			format.format.field = V4L2_FIELD_NONE;
-
-			ret = v4l2_subdev_call(&bru->entity.subdev, pad,
-					       set_fmt, NULL, &format);
-			if (ret < 0)
-				return ret;
-
-			dev_dbg(vsp1->dev,
-				"%s: set format %ux%u (%x) on BRU pad %u\n",
-				__func__, format.format.width,
-				format.format.height,
-				format.format.code, i);
-		}
-		format.pad = bru->entity.source_pad;
+		dev_dbg(vsp1->dev, "%s: set format %ux%u (%x) on %s pad %u\n",
+			__func__, format.format.width, format.format.height,
+			format.format.code, bru_name, i);
 	}
 
+	format.pad = pipe->bru->source_pad;
 	format.format.width = cfg->width;
 	format.format.height = cfg->height;
 	format.format.code = MEDIA_BUS_FMT_ARGB8888_1X32;
 	format.format.field = V4L2_FIELD_NONE;
 
-	if (lif_index == 1) {
-		ret = v4l2_subdev_call(&brs->entity.subdev, pad, set_fmt, NULL,
-				       &format);
-		if (ret < 0)
-			return ret;
+	ret = v4l2_subdev_call(&pipe->bru->subdev, pad, set_fmt, NULL,
+			       &format);
+	if (ret < 0)
+		return ret;
 
-		dev_dbg(vsp1->dev, "%s: set format %ux%u (%x) on BRS pad %u\n",
-			__func__, format.format.width, format.format.height,
-			format.format.code, i);
-	} else {
-		ret = v4l2_subdev_call(&bru->entity.subdev, pad, set_fmt, NULL,
-				       &format);
-		if (ret < 0)
-			return ret;
-
-		dev_dbg(vsp1->dev, "%s: set format %ux%u (%x) on BRU pad %u\n",
-			__func__, format.format.width, format.format.height,
-			format.format.code, i);
-	}
+	dev_dbg(vsp1->dev, "%s: set format %ux%u (%x) on %s pad %u\n",
+		__func__, format.format.width, format.format.height,
+		format.format.code, bru_name, i);
 
 	format.pad = RWPF_PAD_SINK;
-	ret = v4l2_subdev_call(&vsp1->wpf[lif_index]->entity.subdev,
-			       pad, set_fmt, NULL, &format);
+	ret = v4l2_subdev_call(&pipe->output->entity.subdev, pad, set_fmt, NULL,
+			       &format);
 	if (ret < 0)
 		return ret;
 
-	dev_dbg(vsp1->dev, "%s: set format %ux%u (%x) on WPF%d sink\n",
+	dev_dbg(vsp1->dev, "%s: set format %ux%u (%x) on WPF%u sink\n",
 		__func__, format.format.width, format.format.height,
-		format.format.code, lif_index);
+		format.format.code, pipe->output->entity.index);
 
 	format.pad = RWPF_PAD_SOURCE;
-	ret = v4l2_subdev_call(&vsp1->wpf[lif_index]->entity.subdev,
-			       pad, get_fmt, NULL, &format);
+	ret = v4l2_subdev_call(&pipe->output->entity.subdev, pad, get_fmt, NULL,
+			       &format);
 	if (ret < 0)
 		return ret;
 
-	dev_dbg(vsp1->dev, "%s: got format %ux%u (%x) on WPF%d source\n",
+	dev_dbg(vsp1->dev, "%s: got format %ux%u (%x) on WPF%u source\n",
 		__func__, format.format.width, format.format.height,
-		format.format.code, lif_index);
+		format.format.code, pipe->output->entity.index);
 
 	format.pad = LIF_PAD_SINK;
-	ret = v4l2_subdev_call(&vsp1->lif[lif_index]->entity.subdev,
-			       pad, set_fmt, NULL, &format);
+	ret = v4l2_subdev_call(&pipe->lif->subdev, pad, set_fmt, NULL,
+			       &format);
 	if (ret < 0)
 		return ret;
 
-	dev_dbg(vsp1->dev, "%s: set format %ux%u (%x) on LIF%d sink\n",
+	dev_dbg(vsp1->dev, "%s: set format %ux%u (%x) on LIF%u sink\n",
 		__func__, format.format.width, format.format.height,
-		format.format.code, lif_index);
+		format.format.code, pipe_index);
 
-	/* Verify that the format at the output of the pipeline matches the
+	/*
+	 * Verify that the format at the output of the pipeline matches the
 	 * requested frame size and media bus code.
 	 */
 	if (format.format.width != cfg->width ||
@@ -283,13 +268,14 @@ int vsp1_du_setup_lif(struct device *dev, const struct vsp1_du_lif_config *cfg,
 		return -EPIPE;
 	}
 
-	/* Mark the pipeline as streaming and enable the VSP1. This will store
+	/*
+	 * Mark the pipeline as streaming and enable the VSP1. This will store
 	 * the pipeline pointer in all entities, which the s_stream handlers
 	 * will need. We don't start the entities themselves right at this point
 	 * as there's no plane configured yet, so we can't start processing
 	 * buffers.
 	 */
-	ret = vsp1_device_get(vsp1, lif_index);
+	ret = vsp1_device_get(vsp1);
 	if (ret < 0)
 		return ret;
 
@@ -297,11 +283,10 @@ int vsp1_du_setup_lif(struct device *dev, const struct vsp1_du_lif_config *cfg,
 	 * Register a callback to allow us to notify the DRM driver of frame
 	 * completion events.
 	 */
+	drm_pipe->du_complete = cfg->callback;
+	drm_pipe->du_private = cfg->callback_data;
 
-	vsp1->drm->du_complete[lif_index] = cfg->callback;
-	vsp1->drm->du_private[lif_index] = cfg->callback_data;
-
-	ret = media_entity_pipeline_start(&pipe->output->entity.subdev.entity,
+	ret = media_pipeline_start(&pipe->output->entity.subdev.entity,
 					  &pipe->pipe);
 	if (ret < 0) {
 		dev_dbg(vsp1->dev, "%s: pipeline start failed\n", __func__);
@@ -309,6 +294,10 @@ int vsp1_du_setup_lif(struct device *dev, const struct vsp1_du_lif_config *cfg,
 		return ret;
 	}
 
+	/* Disable the display interrupts. */
+	vsp1_write(vsp1, VI6_DISP_IRQ_STA(pipe_index), 0);
+	vsp1_write(vsp1, VI6_DISP_IRQ_ENB(pipe_index), 0);
+
 	dev_dbg(vsp1->dev, "%s: pipeline enabled\n", __func__);
 
 	return 0;
@@ -318,22 +307,21 @@ EXPORT_SYMBOL_GPL(vsp1_du_setup_lif);
 /**
  * vsp1_du_atomic_begin - Prepare for an atomic update
  * @dev: the VSP device
+ * @pipe_index: the DRM pipeline index
  */
-void vsp1_du_atomic_begin(struct device *dev, unsigned int lif_index)
+void vsp1_du_atomic_begin(struct device *dev, unsigned int pipe_index)
 {
 	struct vsp1_device *vsp1 = dev_get_drvdata(dev);
-	struct vsp1_pipeline *pipe = &vsp1->drm->pipe[lif_index];
+	struct vsp1_drm_pipeline *drm_pipe = &vsp1->drm->pipe[pipe_index];
 
-	vsp1->drm->num_inputs[lif_index] = pipe->num_inputs;
-
-	/* Prepare the display list. */
-	pipe->dl = vsp1_dl_list_get(pipe->output->dlm);
+	drm_pipe->enabled = drm_pipe->pipe.num_inputs != 0;
 }
 EXPORT_SYMBOL_GPL(vsp1_du_atomic_begin);
 
 /**
  * vsp1_du_atomic_update - Setup one RPF input of the VSP pipeline
  * @dev: the VSP device
+ * @pipe_index: the DRM pipeline index
  * @rpf_index: index of the RPF to setup (0-based)
  * @cfg: the RPF configuration
  *
@@ -360,10 +348,12 @@ EXPORT_SYMBOL_GPL(vsp1_du_atomic_begin);
  *
  * Return 0 on success or a negative error code on failure.
  */
-int vsp1_du_atomic_update(struct device *dev, unsigned int rpf_index,
+int vsp1_du_atomic_update(struct device *dev, unsigned int pipe_index,
+			  unsigned int rpf_index,
 			  const struct vsp1_du_atomic_config *cfg)
 {
 	struct vsp1_device *vsp1 = dev_get_drvdata(dev);
+	struct vsp1_drm_pipeline *drm_pipe = &vsp1->drm->pipe[pipe_index];
 	const struct vsp1_format_info *fmtinfo;
 	struct vsp1_rwpf *rpf;
 
@@ -376,7 +366,12 @@ int vsp1_du_atomic_update(struct device *dev, unsigned int rpf_index,
 		dev_dbg(vsp1->dev, "%s: RPF%u: disable requested\n", __func__,
 			rpf_index);
 
-		vsp1->drm->inputs[rpf_index].enabled = false;
+		/*
+		 * Remove the RPF from the pipe's inputs. The atomic flush
+		 * handler will disable the input and remove the entity from the
+		 * pipe's entities list.
+		 */
+		drm_pipe->pipe.inputs[rpf_index] = NULL;
 		return 0;
 	}
 
@@ -402,18 +397,21 @@ int vsp1_du_atomic_update(struct device *dev, unsigned int rpf_index,
 	rpf->fmtinfo = fmtinfo;
 	rpf->format.num_planes = fmtinfo->planes;
 	rpf->format.plane_fmt[0].bytesperline = cfg->pitch;
-	if ((rpf->fmtinfo->fourcc == V4L2_PIX_FMT_YUV420M) ||
-		(rpf->fmtinfo->fourcc == V4L2_PIX_FMT_YVU420M) ||
-		(rpf->fmtinfo->fourcc == V4L2_PIX_FMT_YUV422M) ||
-		(rpf->fmtinfo->fourcc == V4L2_PIX_FMT_YVU422M))
+	if (rpf->fmtinfo->fourcc == V4L2_PIX_FMT_YUV420M ||
+	    rpf->fmtinfo->fourcc == V4L2_PIX_FMT_YVU420M ||
+	    rpf->fmtinfo->fourcc == V4L2_PIX_FMT_YUV422M ||
+	    rpf->fmtinfo->fourcc == V4L2_PIX_FMT_YVU422M)
 		rpf->format.plane_fmt[1].bytesperline = cfg->pitch / 2;
 	else
 		rpf->format.plane_fmt[1].bytesperline = cfg->pitch;
 	rpf->alpha = cfg->alpha;
+	rpf->colorkey = cfg->colorkey;
+	rpf->colorkey_en = cfg->colorkey_en;
+	rpf->colorkey_alpha = cfg->colorkey_alpha;
 	rpf->interlaced = cfg->interlaced;
 
 	if ((vsp1->ths_quirks & VSP1_AUTO_FLD_NOT_SUPPORT) &&
-		rpf->interlaced) {
+	    rpf->interlaced) {
 		dev_err(vsp1->dev,
 			"Interlaced mode is not supported.\n");
 		return -EINVAL;
@@ -426,23 +424,24 @@ int vsp1_du_atomic_update(struct device *dev, unsigned int rpf_index,
 	vsp1->drm->inputs[rpf_index].crop = cfg->src;
 	vsp1->drm->inputs[rpf_index].compose = cfg->dst;
 	vsp1->drm->inputs[rpf_index].zpos = cfg->zpos;
-	vsp1->drm->inputs[rpf_index].enabled = true;
+
+	drm_pipe->pipe.inputs[rpf_index] = rpf;
 
 	return 0;
 }
 EXPORT_SYMBOL_GPL(vsp1_du_atomic_update);
 
 static int vsp1_du_setup_rpf_pipe(struct vsp1_device *vsp1,
+				  struct vsp1_pipeline *pipe,
 				  struct vsp1_rwpf *rpf, unsigned int bru_input)
 {
 	struct v4l2_subdev_selection sel;
 	struct v4l2_subdev_format format;
 	const struct v4l2_rect *crop;
-	struct v4l2_subdev *subdev;
 	int ret;
-	bool brs_use = false;
 
-	/* Configure the format on the RPF sink pad and propagate it up to the
+	/*
+	 * Configure the format on the RPF sink pad and propagate it up to the
 	 * BRU sink pad.
 	 */
 	crop = &vsp1->drm->inputs[rpf->entity.index].crop;
@@ -481,7 +480,8 @@ static int vsp1_du_setup_rpf_pipe(struct vsp1_device *vsp1,
 		__func__, sel.r.left, sel.r.top, sel.r.width, sel.r.height,
 		rpf->entity.index);
 
-	/* RPF source, hardcode the format to ARGB8888 to turn on format
+	/*
+	 * RPF source, hardcode the format to ARGB8888 to turn on format
 	 * conversion if needed.
 	 */
 	format.pad = RWPF_PAD_SOURCE;
@@ -506,40 +506,28 @@ static int vsp1_du_setup_rpf_pipe(struct vsp1_device *vsp1,
 	/* BRU sink, propagate the format from the RPF source. */
 	format.pad = bru_input;
 
-	if (bru_input >=
-		(vsp1->info->rpf_count - vsp1->num_brs_inputs))
-		brs_use = true;
-
-	if (vsp1_gen3_vspdl_check(vsp1)) {
-		if (brs_use)
-			subdev = &vsp1->brs->entity.subdev;
-		else
-			subdev = &vsp1->bru->entity.subdev;
-	} else
-		subdev = &vsp1->bru->entity.subdev;
-
-	ret = v4l2_subdev_call(subdev, pad, set_fmt, NULL,
+	ret = v4l2_subdev_call(&pipe->bru->subdev, pad, set_fmt, NULL,
 			       &format);
 	if (ret < 0)
 		return ret;
 
-	dev_dbg(vsp1->dev, "%s: set format %ux%u (%x) on %s pad %u\n",
+	dev_dbg(vsp1->dev, "%s: set format %ux%u (%x) on BRU pad %u\n",
 		__func__, format.format.width, format.format.height,
-		format.format.code, (brs_use ? "BRS" : "BRU"), format.pad);
+		format.format.code, format.pad);
 
 	sel.pad = bru_input;
 	sel.target = V4L2_SEL_TGT_COMPOSE;
 	sel.r = vsp1->drm->inputs[rpf->entity.index].compose;
 
-	ret = v4l2_subdev_call(subdev, pad, set_selection,
-			       NULL, &sel);
+	ret = v4l2_subdev_call(&pipe->bru->subdev, pad, set_selection, NULL,
+			       &sel);
 	if (ret < 0)
 		return ret;
 
 	dev_dbg(vsp1->dev,
-		"%s: set selection (%u,%u)/%ux%u on %s pad %u\n",
+		"%s: set selection (%u,%u)/%ux%u on BRU pad %u\n",
 		__func__, sel.r.left, sel.r.top, sel.r.width, sel.r.height,
-		(brs_use ? "BRS" : "BRU"), sel.pad);
+		sel.pad);
 
 	return 0;
 }
@@ -552,139 +540,123 @@ static unsigned int rpf_zpos(struct vsp1_device *vsp1, struct vsp1_rwpf *rpf)
 /**
  * vsp1_du_atomic_flush - Commit an atomic update
  * @dev: the VSP device
+ * @pipe_index: the DRM pipeline index
  */
-void vsp1_du_atomic_flush(struct device *dev, unsigned int lif_index)
+void vsp1_du_atomic_flush(struct device *dev, unsigned int pipe_index)
 {
 	struct vsp1_device *vsp1 = dev_get_drvdata(dev);
-	struct vsp1_pipeline *pipe = &vsp1->drm->pipe[lif_index];
+	struct vsp1_drm_pipeline *drm_pipe = &vsp1->drm->pipe[pipe_index];
+	struct vsp1_pipeline *pipe = &drm_pipe->pipe;
 	struct vsp1_rwpf *inputs[VSP1_MAX_RPF] = { NULL, };
+	struct vsp1_bru *bru = to_bru(&pipe->bru->subdev);
 	struct vsp1_entity *entity;
+	struct vsp1_entity *next;
+	struct vsp1_dl_list *dl;
+	struct vsp1_dl_body *dlb;
+	const char *bru_name;
 	unsigned long flags;
 	unsigned int i;
-	unsigned int init_rpf, end_rpf;
 	int ret;
 
+	bru_name = pipe->bru->type == VSP1_ENTITY_BRU ? "BRU" : "BRS";
+
+	/* Prepare the display list. */
+	dl = vsp1_dl_list_get(pipe->output->dlm);
+
+	/* Retrieve the default DLB from the list */
+	dlb = vsp1_dl_list_body(dl);
+
 	/* Count the number of enabled inputs and sort them by Z-order. */
 	pipe->num_inputs = 0;
 
-	if (vsp1_gen3_vspdl_check(vsp1)) {
-		if (lif_index == 1) {
-			init_rpf = vsp1->info->rpf_count -
-				       vsp1->num_brs_inputs;
-			end_rpf = vsp1->info->rpf_count;
-		} else {
-			init_rpf = 0;
-			end_rpf = vsp1->info->rpf_count -
-				      vsp1->num_brs_inputs;
-		}
-	} else {
-		init_rpf = 0;
-		end_rpf = vsp1->info->rpf_count;
-	}
-
-	for (i = init_rpf; i < end_rpf; ++i) {
+	for (i = 0; i < vsp1->info->rpf_count; ++i) {
 		struct vsp1_rwpf *rpf = vsp1->rpf[i];
 		unsigned int j;
 
-		if (!vsp1->drm->inputs[i].enabled) {
+		/*
+		 * Make sure we don't accept more inputs than the hardware can
+		 * handle. This is a temporary fix to avoid display stall, we
+		 * need to instead allocate the BRU or BRS to display pipelines
+		 * dynamically based on the number of planes they each use.
+		 */
+		if (pipe->num_inputs >= pipe->bru->source_pad)
 			pipe->inputs[i] = NULL;
-			continue;
-		}
 
-		pipe->inputs[i] = rpf;
+		if (!pipe->inputs[i])
+			continue;
 
 		/* Insert the RPF in the sorted RPFs array. */
 		for (j = pipe->num_inputs++; j > 0; --j) {
-			if (rpf_zpos(vsp1, inputs[j-1+init_rpf]) <=
-					rpf_zpos(vsp1, rpf))
+			if (rpf_zpos(vsp1, inputs[j-1]) <= rpf_zpos(vsp1, rpf))
 				break;
-			inputs[j+init_rpf] = inputs[j-1+init_rpf];
+			inputs[j] = inputs[j-1];
 		}
 
-		inputs[j+init_rpf] = rpf;
+		inputs[j] = rpf;
 	}
 
-	if (!vsp1_gen3_vspdl_check(vsp1))
-		end_rpf = vsp1->info->num_bru_inputs;
+	if (pipe_index == 1 && pipe->num_inputs == 3) {
+		dev_warn(vsp1->dev, "%s: over max RPF's number in %s\n",
+			 __func__, bru_name);
+		return;
+	}
 
 	/* Setup the RPF input pipeline for every enabled input. */
-	for (i = init_rpf; i < end_rpf; ++i) {
+	for (i = 0; i < pipe->bru->source_pad; ++i) {
 		struct vsp1_rwpf *rpf = inputs[i];
-		bool brs_use = false;
 
 		if (!rpf) {
-			vsp1->bru->inputs[i].rpf = NULL;
-			if ((lif_index == 1) && (vsp1->brs))
-				vsp1->brs->inputs[i].rpf = NULL;
+			bru->inputs[i].rpf = NULL;
 			continue;
 		}
 
-		if ((lif_index == 1) && (vsp1->brs)) {
-			vsp1->brs->inputs[i].rpf = rpf;
-			rpf->brs_input = i;
-			brs_use = true;
-		} else {
-			vsp1->bru->inputs[i].rpf = rpf;
-			rpf->bru_input = i;
-		}
+		if (list_empty(&rpf->entity.list_pipe))
+			list_add_tail(&rpf->entity.list_pipe, &pipe->entities);
+
+		bru->inputs[i].rpf = rpf;
+		rpf->bru_input = i;
+		rpf->entity.sink = pipe->bru;
 		rpf->entity.sink_pad = i;
 
 		dev_dbg(vsp1->dev, "%s: connecting RPF.%u to %s:%u\n",
-			__func__, rpf->entity.index,
-			(brs_use ? "BRS" : "BRU"), i);
+			__func__, rpf->entity.index, bru_name, i);
 
-		ret = vsp1_du_setup_rpf_pipe(vsp1, rpf, i);
+		ret = vsp1_du_setup_rpf_pipe(vsp1, pipe, rpf, i);
 		if (ret < 0)
 			dev_err(vsp1->dev,
 				"%s: failed to setup RPF.%u\n",
 				__func__, rpf->entity.index);
 	}
 
-	/*
-	 * If we have a writeback node attached, we use this opportunity to
-	 * update the video buffers.
-	 */
-	if (pipe->output->video && pipe->output->video->frame_end)
-		pipe->output->video->frame_end(pipe);
-
 	/* Configure all entities in the pipeline. */
-	list_for_each_entry(entity, &pipe->entities, list_pipe) {
+	list_for_each_entry_safe(entity, next, &pipe->entities, list_pipe) {
 		/* Disconnect unused RPFs from the pipeline. */
-		if (entity->type == VSP1_ENTITY_RPF) {
-			struct vsp1_rwpf *rpf = to_rwpf(&entity->subdev);
+		if (entity->type == VSP1_ENTITY_RPF &&
+		    !pipe->inputs[entity->index]) {
+			vsp1_dl_fragment_write(dlb, entity->route->reg,
+					       VI6_DPR_NODE_UNUSED);
 
-			if (!pipe->inputs[rpf->entity.index]) {
-				vsp1_dl_list_write(pipe->dl, entity->route->reg,
-						   VI6_DPR_NODE_UNUSED);
-				continue;
-			}
+			list_del_init(&entity->list_pipe);
+
+			continue;
 		}
 
-		vsp1_entity_route_setup(entity, pipe, pipe->dl);
-
-		if (entity->ops->configure) {
-			entity->ops->configure(entity, pipe, pipe->dl,
-					       VSP1_ENTITY_PARAMS_INIT);
-			entity->ops->configure(entity, pipe, pipe->dl,
-					       VSP1_ENTITY_PARAMS_RUNTIME);
-			entity->ops->configure(entity, pipe, pipe->dl,
-					       VSP1_ENTITY_PARAMS_PARTITION);
-		}
+		vsp1_entity_route_setup(entity, pipe, dlb);
+		vsp1_entity_prepare(entity, pipe, dlb);
+		vsp1_entity_configure(entity, pipe, dl, dlb, 0);
 	}
 
-	vsp1_dl_list_commit(pipe->dl, lif_index);
-	pipe->dl = NULL;
+	vsp1_dl_list_commit(dl, pipe_index);
 
 	/* Start or stop the pipeline if needed. */
-	if (!vsp1->drm->num_inputs[lif_index] && pipe->num_inputs) {
-		vsp1_write(vsp1, VI6_DISP_IRQ_STA(lif_index), 0);
-		vsp1_write(vsp1, VI6_DISP_IRQ_ENB(lif_index),
-					 VI6_DISP_IRQ_ENB_DSTE);
+	if (!drm_pipe->enabled && pipe->num_inputs) {
+		vsp1_write(vsp1, VI6_DISP_IRQ_STA((pipe_index)), 0);
+		vsp1_write(vsp1, VI6_DISP_IRQ_ENB((pipe_index)),
+			   VI6_DISP_IRQ_ENB_DSTE);
 		spin_lock_irqsave(&pipe->irqlock, flags);
 		vsp1_pipeline_run(pipe);
 		spin_unlock_irqrestore(&pipe->irqlock, flags);
-	} else if (vsp1->drm->num_inputs[lif_index] && !pipe->num_inputs) {
-		vsp1_write(vsp1, VI6_DISP_IRQ_ENB(lif_index), 0);
+	} else if (drm_pipe->enabled && !pipe->num_inputs) {
 		vsp1_pipeline_stop(pipe);
 	}
 }
@@ -693,41 +665,36 @@ EXPORT_SYMBOL_GPL(vsp1_du_atomic_flush);
 int vsp1_du_map_sg(struct device *dev, struct sg_table *sgt)
 {
 	struct vsp1_device *vsp1 = dev_get_drvdata(dev);
-	struct device *map_dev;
 
-	map_dev = vsp1->fcp ? rcar_fcp_get_device(vsp1->fcp) : dev;
-
-	return dma_map_sg(map_dev, sgt->sgl, sgt->nents, DMA_TO_DEVICE);
+	/*
+	 * As all the buffers allocated by the DU driver are coherent, we can
+	 * skip cache sync. This will need to be revisited when support for
+	 * non-coherent buffers will be added to the DU driver.
+	 */
+	return dma_map_sg_attrs(vsp1->bus_master, sgt->sgl, sgt->nents,
+				DMA_TO_DEVICE, DMA_ATTR_SKIP_CPU_SYNC);
 }
 EXPORT_SYMBOL_GPL(vsp1_du_map_sg);
 
 void vsp1_du_unmap_sg(struct device *dev, struct sg_table *sgt)
 {
 	struct vsp1_device *vsp1 = dev_get_drvdata(dev);
-	struct device *map_dev;
 
-	map_dev = vsp1->fcp ? rcar_fcp_get_device(vsp1->fcp) : dev;
-
-	dma_unmap_sg(map_dev, sgt->sgl, sgt->nents, DMA_TO_DEVICE);
+	dma_unmap_sg_attrs(vsp1->bus_master, sgt->sgl, sgt->nents,
+			   DMA_TO_DEVICE, DMA_ATTR_SKIP_CPU_SYNC);
 }
 EXPORT_SYMBOL_GPL(vsp1_du_unmap_sg);
 
 int vsp1_du_setup_wb(struct device *dev, u32 pixelformat, unsigned int pitch,
-		      dma_addr_t mem[2], unsigned int lif_index)
+		     dma_addr_t mem[2], unsigned int pipe_index)
 {
 	struct vsp1_device *vsp1 = dev_get_drvdata(dev);
-	struct vsp1_pipeline *pipe = &vsp1->drm->pipe[lif_index];
+	struct vsp1_drm_pipeline *drm_pipe = &vsp1->drm->pipe[pipe_index];
+	struct vsp1_pipeline *pipe = &drm_pipe->pipe;
 	struct vsp1_rwpf *wpf = pipe->output;
 	const struct vsp1_format_info *fmtinfo;
-	struct vsp1_rwpf *rpf;
-	unsigned long flags;
+	bool interlaced = false;
 	int i;
-	u32 rpf_num = 0;
-
-	if (pipe->brs)
-		rpf_num = vsp1->info->rpf_count - vsp1->num_brs_inputs;
-
-	rpf = pipe->inputs[rpf_num];
 
 	fmtinfo = vsp1_get_format_info(vsp1, pixelformat);
 	if (!fmtinfo) {
@@ -736,13 +703,18 @@ int vsp1_du_setup_wb(struct device *dev, u32 pixelformat, unsigned int pitch,
 		return -EINVAL;
 	}
 
-	if (rpf->interlaced) {
+	for (i = 0; i < vsp1->info->rpf_count; ++i) {
+		if (!pipe->inputs[i])
+			continue;
+
+		interlaced = pipe->inputs[i]->interlaced;
+	}
+
+	if (interlaced) {
 		dev_err(vsp1->dev, "Prohibited in interlaced mode\n");
 		return -EINVAL;
 	}
 
-	spin_lock_irqsave(&pipe->irqlock, flags);
-
 	wpf->fmtinfo = fmtinfo;
 	wpf->format.num_planes = fmtinfo->planes;
 	wpf->format.plane_fmt[0].bytesperline = pitch;
@@ -751,21 +723,29 @@ int vsp1_du_setup_wb(struct device *dev, u32 pixelformat, unsigned int pitch,
 	for (i = 0; i < wpf->format.num_planes; ++i)
 		wpf->buf_addr[i] = mem[i];
 
-	pipe->output->write_back = 3;
-
-	spin_unlock_irqrestore(&pipe->irqlock, flags);
+	pipe->output->write_back = WB_STAT_CATP_REQUEST;
 
 	return 0;
 }
 EXPORT_SYMBOL_GPL(vsp1_du_setup_wb);
 
-void vsp1_du_wait_wb(struct device *dev, u32 count, unsigned int lif_index)
+int vsp1_du_wait_wb(struct device *dev, u32 count, unsigned int pipe_index)
 {
 	struct vsp1_device *vsp1 = dev_get_drvdata(dev);
-	struct vsp1_pipeline *pipe = &vsp1->drm->pipe[lif_index];
+	struct vsp1_drm_pipeline *drm_pipe = &vsp1->drm->pipe[pipe_index];
+	struct vsp1_pipeline *pipe = &drm_pipe->pipe;
+	int tmp_wb;
 
-	wait_event_interruptible(pipe->event_wait,
-				       (pipe->output->write_back == count));
+	wait_event_interruptible_timeout(pipe->event_wait,
+					 ((tmp_wb = pipe->output->write_back)
+					 <= count), HZ / 10);
+
+	if (tmp_wb != count) {
+		dev_dbg(vsp1->dev,
+			"State transition fail, because high load.\n");
+		return -1;
+	}
+	return 0;
 }
 EXPORT_SYMBOL_GPL(vsp1_du_wait_wb);
 
@@ -773,169 +753,48 @@ EXPORT_SYMBOL_GPL(vsp1_du_wait_wb);
  * Initialization
  */
 
-int vsp1_drm_create_links(struct vsp1_device *vsp1)
-{
-	const u32 flags = MEDIA_LNK_FL_ENABLED | MEDIA_LNK_FL_IMMUTABLE;
-	unsigned int i;
-	unsigned int init_bru_num, end_bru_num;
-	unsigned int init_brs_num, end_brs_num;
-	int ret;
-
-	/* VSPD instances require a BRU to perform composition and a LIF to
-	 * output to the DU.
-	 */
-	if (!vsp1->bru || !vsp1->lif[0])
-		return -ENXIO;
-
-	if (vsp1_gen3_vspdl_check(vsp1)) {
-		if (!vsp1->brs || !vsp1->lif[1])
-			return -ENXIO;
-
-		init_bru_num = 0;
-		init_brs_num = vsp1->info->rpf_count -
-			       vsp1->num_brs_inputs;
-		end_bru_num = vsp1->info->rpf_count -
-			      vsp1->num_brs_inputs;
-		end_brs_num = vsp1->info->rpf_count;
-	} else {
-		init_bru_num = 0;
-		init_brs_num = 0;
-		end_bru_num = vsp1->info->rpf_count;
-		end_brs_num = 0;
-	}
-
-	for (i = init_bru_num; i < end_bru_num; ++i) {
-		struct vsp1_rwpf *rpf = vsp1->rpf[i];
-
-		ret = media_create_pad_link(&rpf->entity.subdev.entity,
-					    RWPF_PAD_SOURCE,
-					    &vsp1->bru->entity.subdev.entity,
-					    i, flags);
-		if (ret < 0)
-			return ret;
-		rpf->entity.sink = &vsp1->bru->entity.subdev.entity;
-		rpf->entity.sink_pad = i;
-	}
-
-	for (i = init_brs_num; i < end_brs_num; ++i) {
-		struct vsp1_rwpf *rpf = vsp1->rpf[i];
-
-		ret = media_create_pad_link(&rpf->entity.subdev.entity,
-					    RWPF_PAD_SOURCE,
-					    &vsp1->brs->entity.subdev.entity,
-					    i, flags);
-		if (ret < 0)
-			return ret;
-		rpf->entity.sink = &vsp1->brs->entity.subdev.entity;
-		rpf->entity.sink_pad = i;
-	}
-
-	ret = media_create_pad_link(&vsp1->bru->entity.subdev.entity,
-				    vsp1->bru->entity.source_pad,
-				    &vsp1->wpf[0]->entity.subdev.entity,
-				    RWPF_PAD_SINK, flags);
-	if (ret < 0)
-		return ret;
-
-	vsp1->bru->entity.sink = &vsp1->wpf[0]->entity.subdev.entity;
-	vsp1->bru->entity.sink_pad = RWPF_PAD_SINK;
-
-	ret = media_create_pad_link(&vsp1->wpf[0]->entity.subdev.entity,
-				    RWPF_PAD_SOURCE,
-				    &vsp1->lif[0]->entity.subdev.entity,
-				    LIF_PAD_SINK, flags);
-	if (ret < 0)
-		return ret;
-
-	if (vsp1_gen3_vspdl_check(vsp1)) {
-		ret = media_create_pad_link(
-					&vsp1->brs->entity.subdev.entity,
-					vsp1->brs->entity.source_pad,
-					&vsp1->wpf[1]->entity.subdev.entity,
-					RWPF_PAD_SINK, flags);
-		if (ret < 0)
-			return ret;
-
-		vsp1->brs->entity.sink = &vsp1->wpf[1]->entity.subdev.entity;
-		vsp1->brs->entity.sink_pad = RWPF_PAD_SINK;
-
-		ret = media_create_pad_link(
-					&vsp1->wpf[1]->entity.subdev.entity,
-					RWPF_PAD_SOURCE,
-					&vsp1->lif[1]->entity.subdev.entity,
-					LIF_PAD_SINK, flags);
-		if (ret < 0)
-			return ret;
-	}
-
-	if (vsp1->wpf[0]->has_writeback) {
-		/* Connect the video device to the WPF for Writeback support */
-		ret = media_create_pad_link(&vsp1->wpf[0]->entity.subdev.entity,
-				    RWPF_PAD_SOURCE,
-				    &vsp1->wpf[0]->video->video.entity,
-				    0, flags);
-		if (ret < 0)
-			return ret;
-	}
-
-	return 0;
-}
-
 int vsp1_drm_init(struct vsp1_device *vsp1)
 {
-	struct vsp1_pipeline *pipe;
-	unsigned int i, j, lif_cnt = 1;
-	int rpf_share = -1;
+	unsigned int i;
 
 	vsp1->drm = devm_kzalloc(vsp1->dev, sizeof(*vsp1->drm), GFP_KERNEL);
 	if (!vsp1->drm)
 		return -ENOMEM;
 
-	if (vsp1_gen3_vspdl_check(vsp1)) {
-		lif_cnt = 2;
-		rpf_share = vsp1->info->rpf_count - vsp1->num_brs_inputs;
-	}
+	/* Create one DRM pipeline per LIF. */
+	for (i = 0; i < vsp1->info->lif_count; ++i) {
+		struct vsp1_drm_pipeline *drm_pipe = &vsp1->drm->pipe[i];
+		struct vsp1_pipeline *pipe = &drm_pipe->pipe;
 
-	for (i = 0; i < lif_cnt; i++) {
-		pipe = &vsp1->drm->pipe[i];
 		vsp1_pipeline_init(pipe);
 
-		if (i == 1) {
-			list_add_tail(&vsp1->brs->entity.list_pipe,
-				      &pipe->entities);
-			pipe->brs = &vsp1->brs->entity;
-		} else {
-			list_add_tail(&vsp1->bru->entity.list_pipe,
-				      &pipe->entities);
-			pipe->bru = &vsp1->bru->entity;
-		}
-		/* The DRM pipeline is static, add entities manually. */
-		for (j = 0; j < vsp1->info->rpf_count; ++j) {
-			struct vsp1_rwpf *input = vsp1->rpf[j];
-
-			if (rpf_share < 0)
-				list_add_tail(&input->entity.list_pipe,
-						 &pipe->entities);
-			else {
-				if ((i == 0) && (j < rpf_share))
-					list_add_tail(&input->entity.list_pipe,
-						      &pipe->entities);
-				if ((i == 1) && (j >= rpf_share))
-					list_add_tail(&input->entity.list_pipe,
-						      &pipe->entities);
-			}
-		}
-		list_add_tail(&vsp1->wpf[i]->entity.list_pipe,
-			      &pipe->entities);
-		list_add_tail(&vsp1->lif[i]->entity.list_pipe,
-			      &pipe->entities);
-
+		/*
+		 * The DRM pipeline is static, add entities manually. The first
+		 * pipeline uses the BRU and the second pipeline the BRS.
+		 */
+		pipe->bru = i == 0 ? &vsp1->bru->entity : &vsp1->brs->entity;
 		pipe->lif = &vsp1->lif[i]->entity;
 		pipe->output = vsp1->wpf[i];
 		pipe->output->pipe = pipe;
-		pipe->output->write_back = 0;
+		pipe->output->write_back = WB_STAT_CATP_DONE;
 		pipe->frame_end = vsp1_du_pipeline_frame_end;
 		init_waitqueue_head(&pipe->event_wait);
+
+		pipe->bru->sink = &pipe->output->entity;
+		pipe->bru->sink_pad = 0;
+		pipe->output->entity.sink = pipe->lif;
+		pipe->output->entity.sink_pad = 0;
+
+		list_add_tail(&pipe->bru->list_pipe, &pipe->entities);
+		list_add_tail(&pipe->lif->list_pipe, &pipe->entities);
+		list_add_tail(&pipe->output->entity.list_pipe, &pipe->entities);
+	}
+
+	/* Disable all RPFs initially. */
+	for (i = 0; i < vsp1->info->rpf_count; ++i) {
+		struct vsp1_rwpf *input = vsp1->rpf[i];
+
+		INIT_LIST_HEAD(&input->entity.list_pipe);
 	}
 
 	return 0;
